/***************************************************************************
 *
 * Copyright (C) 2005 Elad Lahav (elad_lahav@users.sourceforge.net)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ***************************************************************************/

#include <qpainter.h>
#include "graphedge.h"

/**
 * Class constructor.
 * @param	pCanvas	The canvas on which to draw the edge
 * @param	pEdge	The graph edge
 */
GraphEdge::GraphEdge(QCanvas* pCanvas, Agedge_t* pEdge) :
	QCanvasPolygonalItem(pCanvas),
	m_pEdge(pEdge)
{
}

/**
 * Class destructor.
 */
GraphEdge::~GraphEdge()
{
}

/**
 * Determines the coordinates of the polygon part of the edge.
 * @param	arr	A point array describing the polygon
 */
void GraphEdge::setPolyPoints(const QPointArray& arr)
{
	m_arrPoly = arr;
	if (m_arrCurve.size() > 0)
		calcArea();
}

/**
 * Determines the coordinates of the control points used for drawing the
 * spline part of the edge.
 * @param	arr	A point array describing the spline
 */
void GraphEdge::setCurvePoints(const QPointArray& arr)
{
	m_arrCurve = arr;
	if (m_arrPoly.size() > 0)
		calcArea();
}

/**
 * Draws the spline as a sequence of cubic Bezier curves.
 * @param	painter	Used for drawing the item on the canvas view
 */
void GraphEdge::drawShape(QPainter& painter)
{
	uint i;
	
	// Draw the polygon
	painter.drawConvexPolygon(m_arrPoly);
	
	// Draw the Bezier curves
	for (i = 0; i < m_arrCurve.size() - 1; i += 3)
		painter.drawCubicBezier(m_arrCurve, i);
}

/**
 * Calculates the area surrounding the edge, and the bounding rectangle of
 * the edge's polygonal head.
 * The calculated area is used to find items on the canvas and to display
 * tips. The bounding rectangle defines the tip's region (@see QToolTip).
 * TODO: The function assumes that the we can simply append the polygon's
 * array to that of the splines. Is this always the case, or should we sort 
 * the points of the polygon in radial order?
 */
void GraphEdge::calcArea()
{
	uint i;
	int nXpos, nYpos;
	QPoint pt;
	
	// Calculate the edge's area
	m_arrArea.resize(m_arrCurve.size() + m_arrPoly.size() - 1);
	m_arrArea.putPoints(0, m_arrCurve.size(), m_arrCurve);
	m_arrArea.putPoints(m_arrCurve.size(), m_arrPoly.size() - 1, m_arrPoly);
	
	// Calculate the head's bouding rectangle
	m_rcBound = QRect(m_arrPoly[0], m_arrPoly[0]);
	for (i = 1; i < m_arrPoly.size(); i++) {
		nXpos = m_arrPoly[i].x();
		if (nXpos < m_rcBound.left())
			m_rcBound.setLeft(nXpos);
		else if (nXpos > m_rcBound.right())
			m_rcBound.setRight(nXpos);
		
		nYpos = m_arrPoly[i].y();
		if (nYpos < m_rcBound.top())
			m_rcBound.setTop(nYpos);
		else if (nYpos > m_rcBound.bottom())
			m_rcBound.setBottom(nYpos);
	}
}
