/***************************************************************************
 *
 * Copyright (C) 2005 Elad Lahav (elad_lahav@users.sourceforge.net)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ***************************************************************************/

#include <graphwidget.h>
#include <qpaintdevicemetrics.h>

/**
 * The widget on which the graph is drawn.
 * This variable is set by the draw() function.
 */
static GraphWidget* s_pWidget;

/**
 * A graphviz context used for laying out and drawing the graph.
 * This variable is set by the init() function.
 */
static GVC_t* s_pGVC;

#ifdef GRAPHVIZ_2_2

namespace Graphviz {

#define GV_PIXELPERINCH			72

/**
 * Horizontal and vertical DPI for the painting device (the graph widget)
 * Used to convert the 72 DPI measure used by graphviz to the one used by the
 * widget.
 */
static int s_nHorzDpi;
static int s_nVertDpi;

#define INCH_TO_PIXEL_H(_i) 	((_i) * s_nHorzDpi)
#define INCH_TO_PIXEL_V(_i) 	((_i) * s_nVertDpi)
#define INCH_TO_PIXEL_PT(_pt)	\
	QPoint((int)INCH_TO_PIXEL_H(_pt.x), (int)INCH_TO_PIXEL_V(_pt.y))
#define PIXEL_TO_PIXEL_H(_p)	((_p) * s_nHorzDpi / GV_PIXELPERINCH)
#define PIXEL_TO_PIXEL_V(_p)	((_p) * s_nVertDpi / GV_PIXELPERINCH)
#define PIXEL_TO_PIXEL_PT(_pt)	\
	QPoint(PIXEL_TO_PIXEL_H(_pt.x), PIXEL_TO_PIXEL_V(s_box.UR.y - _pt.y))

/**
 * Stores the bounding rectangle of the graph.
 * Used to invert the Y axis of the graph so it goes from top to bottom.
 */
box s_box;
	
#define INIT_DEVICE_METRICS(_w)	{ \
	QPaintDeviceMetrics pdm(_w); \
	s_nHorzDpi = pdm.logicalDpiX(); \
	s_nVertDpi = pdm.logicalDpiY(); \
}

/**
 * Called at the beginning of the drawing process.
 * Resizes the widget to fit the graph's bounding rectangle.
 * @param	br	The graph's bounding rectangle
 */
static void beginGraph(GVC_t*, graph_t*, box br, point)
{
	s_box = br;
	s_pWidget->resize(PIXEL_TO_PIXEL_H(br.UR.x - br.LL.x),
		PIXEL_TO_PIXEL_V(br.UR.y - br.LL.y));
}

/**
 * Called before a node is drawn.
 * @param	pNode	The drawn node
 */
static void beginNode(node_t* pNode)
{
	s_pWidget->startNode(pNode);
}

/**
 * Called after a node has been drawn.
 */
static void endNode()
{
	s_pWidget->endNode();
}

/**
 * Called before an edge is drawn.
 * @param	pEdge	The drawn edge
 */
static void beginEdge(edge_t* pEdge)
{
	s_pWidget->startEdge(pEdge);
}

/**
 * Called after an edge has been drawn.
 */
static void endEdge()
{
	s_pWidget->endEdge();
}

/**
 * Determines the font that should be used for drawing text.
 * @param	szFamily	The font's family name
 * @param	dSize		The font's size
 */
static void setFont(char* szFamily, double dSize)
{
	s_pWidget->setFont(QFont(szFamily, (int)dSize));
}

/**
 * Draws some text at the given position.
 * @param	pt			The point where the text should be drawn
 * @param	pTextLine	Contains the text to draw
 */
static void textLine(point pt, textline_t* pTextLine)
{
	s_pWidget->text(pTextLine->str, PIXEL_TO_PIXEL_PT(pt));
}

/**
 * Draws a polygon using the given points.
 * @param	pArray	An array of points
 * @param	nSize	The size of the array
 */
static void polygon(point* pArray, int nSize, int)
{
	QPointArray arrPoints(nSize + 1);
	int i;
	
	// Create an array of QPoint objects
	for (i = 0; i < nSize; i++)
		arrPoints.setPoint(i, PIXEL_TO_PIXEL_PT(pArray[i]));
	
	// Need to close the shape to get a polygon
	arrPoints.setPoint(nSize, PIXEL_TO_PIXEL_PT(pArray[0]));
	
	// Draw the polygon
	s_pWidget->polygon(arrPoints);
}

/**
 * Draws a spline composed of several cubic Bezier curves using the given 
 * control points.
 * @param	pArray	An array of control points
 * @param	nSize	The size of the array
 */
static void bezierCurve(point* pArray, int nSize, int, int)
{
	QPointArray arrPoints(nSize);
	int i;
	
	// Create an array of QPoint objects
	for (i = 0; i < nSize; i++)
		arrPoints.setPoint(i, PIXEL_TO_PIXEL_PT(pArray[i]));
	
	// Draw the curve
	s_pWidget->curve(arrPoints);
}

/**
 * Call-back functions for drawing the graph based on instructions from
 * graphviz.
 */
static codegen_t s_cg = {
	NULL /* reset */,
	NULL /* begin_job */,
	NULL /* end_job */,
	beginGraph,
	NULL /* end_graph */,
	NULL /* begin_page */,
	NULL /* end_page */ ,
	NULL /* begin_layer */,
	NULL /* end_layer */,
	NULL /* begin_cluster */,
	NULL /* end_cluster */,
	NULL /* begin_nodes */,
	NULL /* end_nodes */,
	NULL /* begin_edges */,
	NULL /* end_edges */,
	beginNode,
	endNode,
	beginEdge,
	endEdge,
	NULL /* begin_context */,
	NULL /* end_context */,
	NULL /* begin_anchor */,
	NULL /* end_anchor */,
	setFont,
	textLine,
	NULL,
	NULL,
	NULL /* set_style */,
	NULL /* ellipse */,
	polygon,
	bezierCurve,
	NULL /* polyline */,
	false,
	NULL /* comment */,
	NULL /* textsize */,
	NULL /* user_shape */,
	NULL /* userShapeSize */
};

/**
 * Initialises the graphviz library.
 * Creates the context used for laying out and drawing the graph.
 * NOTE: The command line arguments are ignored, but we still need to pass the 
 * variable to dotneato_initialize(), or else the compiler issues a warning.
 * @param	ppArgs	Command line arguments
 */
void init(char** ppArgs)
{
	s_pGVC = gvContext();
	dotneato_initialize(s_pGVC, 0, ppArgs);
}

/**
 * Library clean-up.
 */
void fini()
{
	dotneato_terminate(s_pGVC);
}

/**
 * Lays out the graph and performs the drawing process.
 * @param	pWidget	The widget that should be used for drawing
 */
void draw(GraphWidget* pWidget)
{
	Agraph_t* pGraph;
	
	// Initialisation
	s_pWidget = pWidget;
	pGraph = pWidget->getGraph();
	INIT_DEVICE_METRICS(pWidget);
	
	// Set our own code generator
	s_pGVC->codegen = &s_cg;
	gvBindContext(s_pGVC, pGraph);

	// Lay out and draw
	dot_layout(pGraph);
	emit_graph(s_pGVC, EMIT_SORTED);
	dot_cleanup(pGraph);
}

/**
 * Writes a description of a graph to the given file, using the dot language.
 * @param	pWidget	The graph widget that holds the graph to write
 * @param	pFile	An open file to write to
 */
void save(GraphWidget* pWidget, FILE* pFile)
{
	gvBindContext(s_pGVC, pWidget->getGraph());
	dot_layout(pWidget->getGraph());
	agwrite(pWidget->getGraph(), pFile);
	dot_cleanup(pWidget->getGraph());
}

}

#else

#include <gvcext.h>

// NOTE:
// The following code was taken from gvplugin.h. This file contains some
// code which is not compatible with ISO-C, and therefore causes an error when
// compiled with -pedantic-errors.
#define GVPLUGIN_H
extern "C" {

/*
 * Define an apis array of name strings using an enumerated api_t as index.
 * The enumerated type is defined here.  The apis array is
 * inititialized in gvplugin.c by redefining ELEM and reinvoking APIS.
 */
#define APIS ELEM(render) ELEM(layout) ELEM(textlayout) ELEM(device) \
	ELEM(usershape)
	
// NOTE:
// Using APIS for the ELEM macro results in a redundant comma at the end
// of the enumeration list, causing 'gcc -pedantic' to fail.
// The value _DUMMY_ELEM_ fixes the pedantic errors problem.
#define ELEM(x) API_##x,
	typedef enum { APIS _DUMMY_ELEM_ = 0 } api_t;
#undef ELEM

typedef struct {
	int id;	/* an id that is only unique within a package 
			of plugins of the same api.
			A codegen id is unique in in the set of codegens.
			A renderer-type such as "png" in the cairo package
			has an id that is different from the "ps" type
			in the same package */
	char *type;	/* a string name, such as "png" or "ps" that
				distinguishes different types withing the same
				api (renderer in this case) */
	int quality;    /* an arbitrary integer used for ordering plugins of
					the same type from different packages */
	void *engine;   /* pointer to the jump table for the plugin */
	void *features; /* pointer to the feature description 
					void* because type varies by api */
} gvplugin_installed_t;

typedef struct {
	api_t api;
	gvplugin_installed_t *types;
} gvplugin_api_t;

typedef struct {
	char *packagename;		/* used when this plugin is builtin and has
							no pathname */
	gvplugin_api_t *apis;
} gvplugin_library_t;

}

// NOTE: 
// The struct gvdevice_callbacks_s contains a field called "delete", which is 
// a reserved keyword in C++.
#define delete delete_graphviz
#include <gvplugin_render.h>
#undef delete

#include <gvcint.h>
	
#define INIT_DEVICE_METRICS(_w)	{ \
	QPaintDeviceMetrics pdm(_w); \
	s_gvfeat.default_dpi = pdm.logicalDpiX(); \
}

enum {
	FORMAT_GRAPH_WIDGET
};

extern "C" {
extern bool gvplugin_install(GVC_t * gvc, api_t api,
	char *typestr, int quality, char *packagename, char *path,
	gvplugin_installed_t * typeptr);
}

namespace Graphviz {
            
/**
 * Called at the beginning of the drawing process.
 * Resizes the widget to fit the graph's bounding rectangle.
 * @param	pJob	The current rendering job
 */
void beginGraph(GVJ_t* pJob, char*)
{
	s_pWidget->resize(pJob->width, pJob->height);
}
            
/**
 * Called before a node is drawn.
 * @param	pJob	The current rendering job
 */
void beginNode(GVJ_t* pJob, char*, long)
{
	s_pWidget->startNode(pJob->gvc->n);
}

/**
 * Called after a node has been drawn.
 */
void endNode(GVJ_t*)
{
	s_pWidget->endNode();
}

/**
 * Called before an edge is drawn.
 * @param	pJob	The current rendering job
 */
void beginEdge(GVJ_t* pJob, char*, bool, char*, long)
{
	s_pWidget->startEdge(pJob->gvc->e);
}

/**
 * Called after an edge has been drawn.
 */
void endEdge(GVJ_t*)
{
	s_pWidget->endEdge();
}

/**
 * Draws some text at the given position.
 * @param	pt			The point where the text should be drawn
 * @param	pTextLine	Contains the text to draw
 */
void textLine(GVJ_t*, pointf pt, textline_t* pText)
{
	s_pWidget->text(pText->str, QPoint((int)pt.x, (int)pt.y));
}

/**
 * Draws a polygon using the given points.
 * @param	pArray	An array of points
 * @param	nSize	The size of the array
 */
void polygon(GVJ_t*, pointf* pPtArray, int nSize, int)
{
	QPointArray arr(nSize);
	int i;
	
	// Create an array of QPoint objects
	for (i = 0; i < nSize; i++)
		arr.setPoint(i, (int)pPtArray[i].x, (int)pPtArray[i].y);
	
	// Draw the polygon
	s_pWidget->polygon(arr);
}

/**
 * Draws a spline composed of several cubic Bezier curves using the given 
 * control points.
 * @param	pArray	An array of control points
 * @param	nSize	The size of the array
 */
void bezierCurve(GVJ_t*, pointf* pPtArray, int nSize, int, int, int)
{
	QPointArray arr(nSize);
	int i;
	
	// Create an array of QPoint objects
	for (i = 0; i < nSize; i++)
		arr.setPoint(i, (int)pPtArray[i].x, (int)pPtArray[i].y);
	
	// Draw the curve
	s_pWidget->curve(arr);
}

/**
 * Call-back functions for drawing the graph based on instructions from
 * graphviz.
 */
static gvrender_engine_t s_gveng = {
	NULL /* begin_job */,
	NULL /* end_job */,
	beginGraph /* begin_graph */,
	NULL /* end_graph */,
	NULL /* begin_page */,
	NULL /* end_page */ ,
	NULL /* begin_layer */,
	NULL /* end_layer */,
	NULL /* begin_cluster */,
	NULL /* end_cluster */,
	NULL /* begin_nodes */,
	NULL /* end_nodes */,
	NULL /* begin_edges */,
	NULL /* end_edges */,
	beginNode,
	endNode,
	beginEdge,
	endEdge,
	NULL /* begin_anchor */,
	NULL /* end_anchor */,
	textLine,
	NULL /* resolve_color */,
	NULL /* ellipse */,
	polygon,
	bezierCurve,
	NULL /* polyline */,
	NULL /* comment */,
	NULL /* user_shape */
};

/**
 * Defines the plugin features, such as orientation, colour modes and DPI.
 */
static gvrender_features_t s_gvfeat = {
	GVRENDER_DOES_TRUECOLOR | GVRENDER_Y_GOES_DOWN,
	5,
	0, // Will be set by draw()
	0,
	0,
	RGBA_BYTE,
	0
};

/**
 * Associate engine and features with a named format.
 */
static gvplugin_installed_t s_types[] = {
	{ FORMAT_GRAPH_WIDGET, (char*)"graph_widget", 0, &s_gveng, &s_gvfeat },
	{ 0, NULL, 0, NULL, NULL }
};

/**
 * Initialises the graphviz library.
 * Creates the context used for laying out and drawing the graph. Installs
 * a built-in plugin that is used for receiveing drawing instructions from
 * graphviz.
 */
void init(char**)
{
	// Create and initialise a new context
	s_pGVC = gvContext();
	
	// Install our plugin
	gvplugin_install(s_pGVC, API_render, (char*)"kscope", 0, (char*)"kscope",
		NULL, s_types);
}

/**
 * Library clean-up.
 * Frees the context created in init().
 */
void fini()
{
	gvFreeContext(s_pGVC);	
}

/**
 * Lays out the graph and performs the drawing process.
 * @param	pWidget	The widget that should be used for drawing
 */
void draw(GraphWidget* pWidget)
{
	Agraph_t* pGraph;
	
	// Initialisation
	s_pWidget = pWidget;
	pGraph = pWidget->getGraph();
	INIT_DEVICE_METRICS(pWidget);
	
	// Lay out and draw
	gvLayout(s_pGVC, pGraph, (char*)"dot");
	gvRender(s_pGVC, pGraph, (char*)"kscope", NULL);
	gvFreeLayout(s_pGVC, pGraph);
}

/**
 * Writes a description of a graph to the given file, using the dot language.
 * @param	pWidget	The graph widget that holds the graph to write
 * @param	pFile	An open file to write to
 */
void save(GraphWidget* pWidget, FILE* pFile)
{
	gvLayout(s_pGVC, pWidget->getGraph(), (char*)"dot");
	agwrite(pWidget->getGraph(), pFile);
	gvFreeLayout(s_pGVC, pWidget->getGraph());
}

}

#endif // GRAPHVIZ_2_2
