/***************************************************************************
 *
 * Copyright (C) 2005 Elad Lahav (elad_lahav@users.sourceforge.net)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ***************************************************************************/

#include <qpainter.h>
#include <qtooltip.h>
#include <klocale.h>
#include "graphwidget.h"
#include "graphnode.h"
#include "graphedge.h"
#include "kscopeconfig.h"
#include "queryviewdlg.h"

#define GRAPH_WIDGET_NAME	(char*)"graphwidget"

#define SET_GRAPH_ATTR(_attr, _val) \
	agxset(m_pGraph, agraphattr(m_pGraph, (char*)_attr, (char*)"")->index, \
		(char*)_val)
#define GET_GRAPH_ATTR(_attr) \
	agxget(m_pGraph, agraphattr(m_pGraph, (char*)_attr, (char*)"")->index)

#define SET_NODE_ATTR(_node, _attr, _val) \
	agxset(_node, agnodeattr(m_pGraph, (char*)_attr, (char*)"")->index, \
		(char*)_val)
#define GET_NODE_ATTR(_node, _attr) \
	agxget(_node, agnodeattr(m_pGraph, (char*)_attr, (char*)"")->index)

#define SET_EDGE_ATTR(_edge, _attr, _val) \
	agxset(_edge, agedgeattr(m_pGraph, (char*)_attr, (char*)"")->index, \
		(char*)_val)
#define GET_EDGE_ATTR(_edge, _attr) \
	agxget(_edge, agedgeattr(m_pGraph, (char*)_attr, (char*)"")->index)

typedef Agedge_t* (*FirstEdgeFunc)(Agraph_t*, Agnode_t*);
typedef Agedge_t* (*NextEdgeFunc)(Agraph_t*, Agedge_t*);

char ATTR_TRUE[] = "true";
char ATTR_FALSE[] = "false";

#define IS_TRUE(_s)	(_s[0] == 't')

const char* GRAPH_DIRS[] = { "TB", "LR", "BT", "RL" };

/**
 * Displays a tool tip on the graph.
 * Note that we cannot use the standard tool tip class here, since graph
 * items are neither rectangular nor is their position known in advance.
 * @author	Elad Lahav
 */
class GraphTip : public QToolTip
{
public:
	/**
	 * Class constructor.
	 * @param	pWidget	Owner graph widget
	 */
	GraphTip(GraphWidget* pWidget) : QToolTip(pWidget->viewport()),
		m_pGraphWidget(pWidget) {}
	
	/**
	 * Class destructor.
	 */
	virtual ~GraphTip() {}
	
protected:
	/**
	 * Called when the pre-conditions for a tool tip are met.
	 * Asks the owner for a tip to display and, if one is returned, shows
	 * the tool tip.
	 * @param	ptPos	Current mouse position
	 */
	virtual void maybeTip(const QPoint& ptPos) {
		QString sText;
		QRect rc;
		
		// Display a tip, if required by the owner
		sText = m_pGraphWidget->getTip(ptPos, rc);
		if (sText != QString::null)
			tip(rc, sText);
	}
	
private:
	/** The parent graph widget. */
	GraphWidget* m_pGraphWidget;
};

/**
 * Provides a menu separator with text.
 * The separator is added with QMenuData::insertItem(QWidget*).
 * @author Elad Lahav
 */
class MenuLabel : public QLabel
{
public:
	/**
	 * Class constructor.
	 * @param	sText	The text to display
	 * @param	pParent	The parent widget
	 */
	MenuLabel(const QString& sText, QWidget* pParent) :
		QLabel(sText, pParent) {
		// Set the appropriate visual properties
		setFrameShape(MenuBarPanel);
		setAlignment(AlignHCenter | AlignVCenter);
		setIndent(0);
	}
};

/**
 * Class constructor.
 * @param	pParent	The parent widget
 * @param	szName	The widget's name
 */
GraphWidget::GraphWidget(QWidget* pParent, const char* szName) :
	QCanvasView(pParent, szName),
	m_pCanvas(NULL),
	m_pGraph(NULL),
	m_progress(this),
	m_pCurItem(NULL),
	m_dZoom(1.0),
	m_nMaxNodeDegree(10) // will be overriden by CallTreeDlg
{
	// Create a persistent Cscope process
	m_pCscope = new CscopeFrontend();
	
	// Add records output by the Cscope process
	connect(m_pCscope, SIGNAL(dataReady(FrontendToken*)), this,
		SLOT(slotDataReady(FrontendToken*)));
	
	// Display query progress information
	connect(m_pCscope, SIGNAL(progress(int, int)), this,
		SLOT(slotProgress(int, int)));	
		
	// Draw the graph when the process has finished
	connect(m_pCscope, SIGNAL(finished(uint)), this,
		SLOT(slotFinished(uint)));
	
	// Show a multi-call node when a query results in too many records
	connect(m_pCscope, SIGNAL(aborted()), this,
		SLOT(slotAborted()));
	
	// Create the node popup menu
	m_pNodePopup = new QPopupMenu(this);
	
	m_pNodePopup->insertItem(new MenuLabel(i18n("<b>Called Functions</b>"),
		m_pNodePopup));
	m_pNodePopup->insertItem(i18n("Show"), this,
		SLOT(slotShowCalled()));
	m_pNodePopup->insertItem(i18n("List/Filter..."), this,
		SLOT(slotListCalled()));
	m_pNodePopup->insertItem(i18n("Hide"), this,
		SLOT(slotHideCalled()));
	
	m_pNodePopup->insertItem(new MenuLabel(i18n("<b>Calling Functions</b>"),
		m_pNodePopup));
	m_pNodePopup->insertItem(i18n("Show"), this,
		SLOT(slotShowCalling()));
	m_pNodePopup->insertItem(i18n("List/Filter..."), this,
		SLOT(slotListCalling()));
	m_pNodePopup->insertItem(i18n("Hide"), this,
		SLOT(slotHideCalling()));
	
	m_pNodePopup->insertItem(new MenuLabel(i18n("<b>This Function</b>"),
		m_pNodePopup));
	m_pNodePopup->insertItem(i18n("Find Definition"), this,
		SLOT(slotFindDef()));
	m_pNodePopup->insertItem(i18n("Remove"), this, SLOT(slotRemoveNode()));
		
	// Create the multi-call node popup menu
	m_pMultiCallPopup = new QPopupMenu(this);
	m_pMultiCallPopup->insertItem(i18n("List..."), this,
		SLOT(slotMultiCallDetails()));
	m_pMultiCallPopup->insertSeparator();
	m_pMultiCallPopup->insertItem(i18n("Remove"), this,
		SLOT(slotRemoveNode()));
	
	// Create the edge menu
	m_pEdgePopup = new QPopupMenu(this);
	m_pEdgePopup->insertItem(i18n("Open Call"), this, SLOT(slotOpenCall()));
		
	(void)new GraphTip(this);
	
	// Create a new graph
	reset();
}

/**
 * Class destructor.
 */
GraphWidget::~GraphWidget()
{
	if (m_pGraph != NULL)
		agclose(m_pGraph);
		
	if (m_pCanvas != NULL)
		delete m_pCanvas;
}

/**
 * Creates a new graphviz graph object.
 * If there was a previous graph, it is deleted.
 */
void GraphWidget::reset()
{
	// Clean up
	if (m_pGraph != NULL)
		agclose(m_pGraph);
		
	// Create a new graph, and set its attributes
	m_pGraph = agopen(GRAPH_WIDGET_NAME, AGDIGRAPH);
	SET_GRAPH_ATTR("bgcolor", 
		Config().getColor(KScopeConfig::GraphBack).name().latin1());
	SET_GRAPH_ATTR("rankdir", Config().getGraphOrientation().latin1());
	SET_GRAPH_ATTR("concentrate", ATTR_TRUE);
	
	// Set default node properties
	setNodeAttrib();
}

/**
 * Creates a root node for the graph.
 * The root node defines the connected component which is always displayed
 * (all other connected components are removed when they are no longer
 * strongly connected to the root).
 * @param	sFunc	The function name for the root node
 */
void GraphWidget::setRoot(const QString& sFunc)
{
	// Insert a new node to the graph
	(void)agnode(m_pGraph, (char*)sFunc.latin1());
	draw();
}

/**
 * Adds a call to the graph.
 * A call is made between two functions, the caller and the callee.
 * @param	data	Contains information on the call
 */
void GraphWidget::addCall(const CallData& data)
{
	Agnode_t* pCaller, * pCallee;
	Agedge_t* pEdge;
	
	// Find the relevant nodes (create new nodes if necessary)
	pCaller = agnode(m_pGraph, (char*)data.m_sCaller.latin1());
	pCallee = agnode(m_pGraph, (char*)data.m_sCallee.latin1());
	
	// Create a directed edge
	pEdge = agedge(m_pGraph, pCaller, pCallee);
	SET_EDGE_ATTR(pEdge, "kscope_file", data.m_sFile.latin1());
	SET_EDGE_ATTR(pEdge, "kscope_line", data.m_sLine.latin1());
	SET_EDGE_ATTR(pEdge, "kscope_text", data.m_sText.latin1());
}

/**
 * Creates a special node representing multiple calls to/from a function.
 * Such a node is creates when the number of calls to/from a function exceeds
 * a certain number. Thus the graph does not become too cluttered.
 * A multiple call node can be replaced by some/all of the actual calls by
 * using the "Details..." action in the node's popup menu.
 * @param	sFunc	The parent function
 * @param	bCalled	true if the multiple calls are called from that function,
 *					false if they are calling the function
 */
void GraphWidget::addMultiCall(const QString& sFunc, bool bCalled)
{
	QString sMulti;
	Agnode_t* pFunc, * pMulti;
	
	// Create a unique name for the new node.
	sMulti = sFunc + "$MULTI_" + (bCalled ? "CALLED" : "CALLING");
	
	// Find the relevant nodes (create new nodes if necessary)
	pFunc = agnode(m_pGraph, (char*)sFunc.latin1());
	pMulti = agnode(m_pGraph, (char*)sMulti.latin1());
	
	// Set properties of the new node
	SET_NODE_ATTR(pMulti, "label", "...");
	SET_NODE_ATTR(pMulti, "fillcolor",
		Config().getColor(KScopeConfig::GraphMultiCall).name().latin1());
	SET_NODE_ATTR(pMulti, "kscope_multi", ATTR_TRUE);
	SET_NODE_ATTR(pMulti, "kscope_multi_parent", sFunc.latin1());
	SET_NODE_ATTR(pMulti, "kscope_multi_called", 
		bCalled ? ATTR_TRUE : ATTR_FALSE);
	
	// Create a directed edge
	if (bCalled)
		agedge(m_pGraph, pFunc, pMulti);
	else	
		agedge(m_pGraph, pMulti, pFunc);
}


/**
 * Draws the graph on the canvas using the graphviz engine.
 * A new canvas is created, so all items need to be regenerated.
 * TODO: Can we use the same canvas and only reposition existing items?
 */
void GraphWidget::draw()
{
	QWMatrix mtx;
		
	// Delete the old canvas
	if (m_pCanvas != NULL)
		delete m_pCanvas;
		
	// Create a new canvas
	m_pCanvas = new QCanvas;
	setCanvas(m_pCanvas);
	
	// Apply the zoom factor
	mtx.scale(m_dZoom, m_dZoom);
	setWorldMatrix(mtx);

	// Draw the graph
	Graphviz::draw(this);
}

/**
 * Stores a graph on a file.
 * The file uses the dot language to describe the graph.
 * @param	pFile	An open file to write to
 */
void GraphWidget::save(FILE* pFile)
{
	// Write the zoom factor
	fprintf(pFile, "Zoom=%lf\n", m_dZoom);
	
	// Write the graph using the dot language
	Graphviz::save(this, pFile);
}

/**
 * Exports the  a graph to a dot file.
 * @param	sFile	The full path of the file to export to
 */
void GraphWidget::save(const QString& sFile)
{
	FILE* pFile;
	
	// Open/create the file
	pFile = fopen(sFile.latin1(), "w+");
	if (pFile == NULL)
		return;
	
	// Write the graph using the dot language
	Graphviz::save(this, pFile);
	
	// Clean up
	fclose(pFile);
}

/**
 * Loads a graph from a file.
 * The file should contain a graph represented in the dot language.
 * @param	pFile	An open file to read from
 * @return	true if successful, false otherwise
 */
bool GraphWidget::load(FILE* pFile)
{
	// Read the zoom factor
	m_dZoom = 1.0;
	if (fscanf(pFile, "Zoom=%lf\n", &m_dZoom) != 1)
		return false;
	
	// Read the graph stored on this file
	m_pGraph = agread(pFile);
	
	if (m_pGraph == NULL)
		return false;
		
	// Set default node attributes
	setNodeAttrib();
	return true;
}

/**
 * Changes the zoom factor.
 * @param	bIn	true to zoom in, false to zoom out
 */
void GraphWidget::zoom(bool bIn)
{
	QWMatrix mtx;
	
	// Set the new zoom factor
	if (bIn)
		m_dZoom *= 2.0;
	else
		m_dZoom /= 2.0;
		
	// Apply the transformation matrix
	mtx.scale(m_dZoom, m_dZoom);
	setWorldMatrix(mtx);
}

/**
 * Changes the graph's direction 90 degrees counter-clockwise.
 */
void GraphWidget::rotate()
{
	QString sDir;
	int i;
	
	// Get the current direction
	sDir = GET_GRAPH_ATTR("rankdir");
	
	// Find the next direction
	for (i = 0; i < 4 && sDir != GRAPH_DIRS[i]; i++);
	if (i == 4)
		i = 0;
	else
		i = (i + 1) % 4;

	// Set the new direction
	sDir = GRAPH_DIRS[i % 4];
	SET_GRAPH_ATTR("rankdir", sDir.latin1());
	Config().setGraphOrientation(sDir);
}

/**
 * Checks if a tool tip is required for the given position.
 * NOTE: We currently return a tool tip for edges only
 * @param	ptPos	The position to query
 * @param	rc		Holds the tip's rectangle, upon return
 * @return	The tip's text, or QString::null if no tip is required
 */
QString GraphWidget::getTip(const QPoint& ptPos, QRect& rc)
{
	QPoint ptRealPos, ptTopLeft, ptBottomRight;
	QCanvasItemList il;
	QCanvasItemList::Iterator itr;
	GraphEdge* pEdge;
	QString sText, sFile, sLine;
	
	ptRealPos = viewportToContents(ptPos);
	ptRealPos /= m_dZoom;
	pEdge = NULL;
	
	// Check if there is an edge at this position
	il = m_pCanvas->collisions(ptRealPos);
	for (itr = il.begin(); itr != il.end(); ++itr) {
		pEdge = dynamic_cast<GraphEdge*>(*itr);
		if (pEdge != NULL)
			break;
	}
	
	// No tip if no edge was found
	if (pEdge == NULL)
		return QString::null;
	
	// Set the rectangle for the tip (the tip is closed when the mouse leaves
	// this area)
	rc = pEdge->boundingRect();
	ptTopLeft = rc.topLeft();
	ptBottomRight = rc.bottomRight();
	ptTopLeft *= m_dZoom;	
	ptBottomRight *= m_dZoom;
	ptTopLeft = contentsToViewport(ptTopLeft);
	ptBottomRight = contentsToViewport(ptBottomRight);
	rc = QRect(ptTopLeft, ptBottomRight);
	
	// Create a tip for this edge
	sText = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_text");
	sFile = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_file");
	sLine = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_line");
			
	sText += "<br><i>" + sFile + "</i>:" + sLine;
	return sText;
}

/**
 * Resizes the canvas.
 * @param	size	The new size to use
 */
void GraphWidget::resize(int nWidth, int nHeight)
{
	m_pCanvas->resize(nWidth + 1, nHeight + 1);
}

/**
 * Called before a node is drawn.
 * Creates a new GraphNode item.
 * @param	pNode	Pointer to the node object
 */
void GraphWidget::startNode(Agnode_t* pNode)
{
	m_pCurItem = new GraphNode(m_pCanvas, pNode);
	m_pCurItem->setZ(2.0);
	m_pCurItem->setPen(QPen(Qt::black));
	m_pCurItem->setBrush(Config().getColor(KScopeConfig::GraphNode));
	((GraphNode*)m_pCurItem)->setFont(Config().getFont(KScopeConfig::Graph));
}

/**
 * Called after a node has been drawn.
 * Displays the newly created GraphNode item.
 */
void GraphWidget::endNode()
{
	m_pCurItem->show();
	m_pCurItem = NULL;
}

/**
 * Called before an edge is drawn.
 * Creates a new ArrowEdge item.
 */
void GraphWidget::startEdge(Agedge_t* pEdge)
{
	m_pCurItem = new GraphEdge(m_pCanvas, pEdge);
	m_pCurItem->setZ(1.0);
	m_pCurItem->setPen(QPen(Qt::black));
	m_pCurItem->setBrush(QBrush(Qt::black));
}

/**
 * Called after an edge has been drawn.
 * Displays the newly created ArrowEdge item.
 */
void GraphWidget::endEdge()
{
	m_pCurItem->show();
	m_pCurItem = NULL;
}

/**
 * Draws a polygon using the given points.
 * A polygon is drawn for either a GraphNode, or for the arrow of an
 * ArrowEdge.
 * @param	arrPoints	Points outlining the polygon
 */
void GraphWidget::polygon(const QPointArray& arrPoints)
{
	GraphNode* pNode;
	GraphEdge* pEdge;
	
	// A NULL item means that graphviz is painting the graph's background.
	// We simply set the background colour of the canvas, based on the current
	// brush.
	if (m_pCurItem == NULL) {
		m_pCanvas->setBackgroundColor(GET_GRAPH_ATTR("bgcolor"));
		return;
	}
	
	// Set the polygon's coordinates
	if ((pNode = dynamic_cast<GraphNode*>(m_pCurItem)) != NULL)
		pNode->setPoints(arrPoints);
	else if ((pEdge = dynamic_cast<GraphEdge*>(m_pCurItem)) != NULL)
		pEdge->setPolyPoints(arrPoints);
}

/**
 * Draws the given text at the given position.
 * Currently text is only drawn inside a GraphNode canvas item.
 * @param	sText	The text to draw
 * @param	ptPos	The position of the text
 */
void GraphWidget::text(const QString& sText, const QPoint& ptPos)
{
	GraphNode* pNode;
	
	if ((pNode = dynamic_cast<GraphNode*>(m_pCurItem)) != NULL) {
		pNode->setText(sText);
		pNode->setTextPos(ptPos);
	}
}

/**
 * Draws a Bezier curve using the given control points.
 * @param	arrPoints	The control points to use
 */
void GraphWidget::curve(const QPointArray& arrPoints)
{
	GraphEdge* pEdge;
	
	pEdge = dynamic_cast<GraphEdge*>(m_pCurItem);
	if (pEdge != NULL) {
		pEdge->setPen(Qt::black);
		pEdge->setCurvePoints(arrPoints);
	}
}

/**
 * Draws the contents of the canvas on this view.
 * NOTE: This method is overriden to fix a strange bug in Qt that leaves
 * a border around the canvas part of the view. It should be deleted once
 * this bug is fixed.
 * TODO: Is there a better way of erasing the border?
 * @param	pPainter	Used to paint on the view
 * @param	nX			The horizontal origin of the area to draw
 * @param	nY			The vertical origin of the area to draw
 * @param	nWidth		The width of the area to draw
 * @param	nHeight		The height of the area to draw
 */
void GraphWidget::drawContents(QPainter* pPainter, int nX, int nY, 
	int nWidth, int nHeight)
{
	// Draw the contents of the canvas
	QCanvasView::drawContents(pPainter, nX, nY, nWidth, nHeight);
	
	// Erase the canvas's area border
	if (m_pCanvas) {
		QRect rect = m_pCanvas->rect();
		pPainter->setBrush(QBrush()); // Null brush
		pPainter->setPen(Config().getColor(KScopeConfig::GraphBack));
		pPainter->drawRect(-1, -1, rect.width() + 2, rect.height() + 2);
	}
}

/**
 * Handles mouse clicks over the graph view.
 * @param	pEvent	Includes information on the mouse press event
 */
void GraphWidget::contentsMousePressEvent(QMouseEvent* pEvent)
{
	QPoint ptRealPos;
	QCanvasItemList il;
	QCanvasItemList::Iterator itr;
	QString sFunc;
	GraphNode* pNode;
	GraphEdge* pEdge;
	
	pNode = NULL;
	pEdge = NULL;
	
	// Handle right-clicks only
	if (pEvent->button() != Qt::RightButton) {
		QCanvasView::contentsMousePressEvent(pEvent);
		return;
	}
	
	// Take the zoom factor into consideration
	ptRealPos = pEvent->pos();
	ptRealPos /= m_dZoom;
	
	// Check if an item was clicked
	il = m_pCanvas->collisions(ptRealPos);
	for (itr = il.begin(); itr != il.end(); ++itr) {
		if (dynamic_cast<GraphNode*>(*itr) != NULL)
			pNode = dynamic_cast<GraphNode*>(*itr);
		else if (dynamic_cast<GraphEdge*>(*itr) != NULL)
			pEdge = dynamic_cast<GraphEdge*>(*itr);
	}
	
	// Handle clicks over different types of items
	if (pNode != NULL) {
		// Show a context menu for nodes
		showNodeMenu(pNode, pEvent->globalPos());
	}
	else if (pEdge != NULL) {
		// Show a context menu for edges
		showEdgeMenu(pEdge, pEvent->globalPos());
	}
	else {
		// Take the default action
		QCanvasView::contentsMousePressEvent(pEvent);
	}
}

/**
 * Sets defsault node attributes for the current graph.
 */
void GraphWidget::setNodeAttrib()
{
	Agnode_t* pNode;
	QFont font;
	
	font = Config().getFont(KScopeConfig::Graph);
	
	pNode = m_pGraph->proto->n;
	SET_NODE_ATTR(pNode, "shape", "box");
	SET_NODE_ATTR(pNode, "height", "0.01");
	SET_NODE_ATTR(pNode, "style", "filled");
	SET_NODE_ATTR(pNode, "fillcolor", 
		Config().getColor(KScopeConfig::GraphNode).name().latin1());
	SET_NODE_ATTR(pNode, "fontcolor", 
		Config().getColor(KScopeConfig::GraphText).name().latin1());
	SET_NODE_ATTR(pNode, "fontname", font.family().latin1());
	SET_NODE_ATTR(pNode, "fontsize",
		QString::number(font.pointSize()).latin1());
	SET_NODE_ATTR(pNode, "kscope_called", ATTR_FALSE);
	SET_NODE_ATTR(pNode, "kscope_calling", ATTR_FALSE);
	SET_NODE_ATTR(pNode, "kscope_multi", ATTR_FALSE);
}

/**
 * Removes all edges attached to a function node at the given direction.
 * Any strongly connected components that are no longer connected to that
 * function are deleted.
 * @param	pGraphNode	The node for which to remove the edges
 * @param	bOut		true for outgoing edges, false for incoming
 */
void GraphWidget::removeEdges(GraphNode* pGraphNode, bool bOut)
{
	FirstEdgeFunc pFirst;
	NextEdgeFunc pNext;
	Agnode_t* pNode;
	Agedge_t* pEdge;
	Agnode_t* pDelNode, * pPrevNode;
	Agraph_t** pSubGraphs;
	int nSubGraphs, i;
	
	pNode = pGraphNode->getNode();
	
	// Set edge functions based on requested direction
	if (bOut) {
		pFirst = agfstout;
		pNext = agnxtout;
	}
	else {
		pFirst = agfstin;
		pNext = agnxtin;
	}
	
	// Nothing to do if no edges need to be removed
	if ((pEdge = pFirst(m_pGraph, pNode)) == NULL)
		return;
	
	// Remove all edges in the requested direction
	for (; pEdge != NULL; pEdge = pNext(m_pGraph, pEdge)) {
		if (bOut)
			agdelete(m_pGraph, pEdge->head);
		else
			agdelete(m_pGraph, pEdge->tail);
	}
	
	// Remove all subgraphs except the one that contains the given function
	pSubGraphs = ccomps(m_pGraph, &nSubGraphs, NULL);
	for (i = 0; i < nSubGraphs; i++) {
		if (!agcontains(pSubGraphs[i], pNode)) {
			// Remove all nodes in the subgraph
			for (pDelNode = agfstnode(pSubGraphs[i]); pDelNode != NULL;) {
				pPrevNode = pDelNode;
				pDelNode = agnxtnode(pSubGraphs[i], pDelNode);
				agdelete(m_pGraph, pPrevNode);
			}
			
			// Delete the subgraph
			agclose(pSubGraphs[i]);
		}
	}
}

/**
 * Shows a popup menu for a node.
 * This menu is shown after a node has been right-clicked.
 * @param	pNode	The node for which to show the menu
 * @param	ptPos	The position of the menu 
 */
void GraphWidget::showNodeMenu(GraphNode* pNode, const QPoint& ptPos)
{
	// Remember the node
	m_pMenuItem = pNode;
	
	// Show the popup menu.
	if (IS_TRUE(GET_NODE_ATTR(pNode->getNode(), "kscope_multi")))
		m_pMultiCallPopup->popup(ptPos);
	else
		m_pNodePopup->popup(ptPos);
}

/**
 * Shows a popup menu for an edge.
 * This menu is shown after an edge has been right-clicked.
 * @param	pEdge	The edge for which to show the menu
 * @param	ptPos	The position of the menu 
 */
void GraphWidget::showEdgeMenu(GraphEdge* pEdge, const QPoint& ptPos)
{
	// Remember the edge
	m_pMenuItem = pEdge;
	
	// Show the popup menu.
	m_pEdgePopup->popup(ptPos);
}

/**
 * Adds an entry to the tree, as the child of the active item.
 * Called by a CscopeFrontend object, when a new entry was received in its
 * whole from the Cscope back-end process. The entry contains the data of a
 * function calling the function described by the active item.
 * @param	pToken	The first token in the entry
 */
void GraphWidget::slotDataReady(FrontendToken* pToken)
{
	CallData data;
	QString sFunc;
	
	// Get the file name
	data.m_sFile = pToken->getData();
	pToken = pToken->getNext();

	// Get the function name
	sFunc = pToken->getData();
	pToken = pToken->getNext();

	// Get the line number (do not accept global information on a call tree)
	data.m_sLine = pToken->getData();
	if (data.m_sLine.toUInt() == 0)
		return;
		
	pToken = pToken->getNext();

	// Get the line's text
	data.m_sText = pToken->getData();
		
	// Determine the caller and the callee
	if (m_bCalled) {
		data.m_sCaller = m_sQueriedFunc;
		data.m_sCallee = sFunc;
	}
	else {
		data.m_sCaller = sFunc;
		data.m_sCallee = m_sQueriedFunc;
	}
		
	// Add the call to the graph
	addCall(data);
}

/**
 * Displays search progress information.
 * This slot is connected to the progress() signal emitted by a
 * CscopeFrontend object.
 * @param	nProgress	The current progress value
 * @param	nTotal		The expected final value
 */
void GraphWidget::slotProgress(int nProgress, int nTotal)
{
	m_progress.setProgress(nProgress, nTotal);
}

/**
 * Disables the expandability feature of an item, if no functions calling it
 * were found.
 * @param	nRecords	Number of records reported by the query
 */
void GraphWidget::slotFinished(uint /*nRecords*/)
{
	// Destroy the progress bar
	m_progress.finished();
		
	// Redraw the graph
	draw();
}

/**
 * Adds a multiple call node when a query results in too many entries.
 * This slot is attached to the aborted() signal of the Cscope process.
 */
void GraphWidget::slotAborted()
{
	addMultiCall(m_sQueriedFunc, m_bCalled);
}

/**
 * Shows functions called from the current function node.
 * This slot is connected to the "Show Called Functions" popup menu
 * action.
 */
void GraphWidget::slotShowCalled()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Run a query for called functions
	m_sQueriedFunc = pNode->getText();
	m_bCalled = true;
	m_pCscope->query(CscopeFrontend::Called, m_sQueriedFunc,
		Config().getGraphMaxNodeDegree());
}

/**
 * Shows a list of function calls from the current node.
 * The list is displayed in a query dialogue. The user can the select which
 * calls should be displayed in the graph.
 * This slot is connected to the "List Called Functions" popup menu
 * action.
 */
void GraphWidget::slotListCalled()
{
	GraphNode* pNode;

	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	QueryViewDlg dlg;
	
	// Show the query view dialogue
	dlg.query(CscopeFrontend::Called, pNode->getText());
	if (dlg.exec() != QDialog::Accepted)
		return;
		
	// The OK button was clicked, replace current calls with the listed ones
	removeEdges(pNode, true);
	
	QueryView::Iterator itr;
	CallData data;
	
	data.m_sCaller = pNode->getText();
	
	// Add all listed calls
	for (itr = dlg.getIterator(); !itr.isEOF(); itr.next()) {
		data.m_sCallee = itr.getFunc();
		data.m_sFile = itr.getFile();
		data.m_sLine = itr.getLine();
		data.m_sText = itr.getText();
		
		addCall(data);
	}
	
	// Redraw the graph
	draw();
}

/**
 * Hides functions called from the current function node.
 * This slot is connected to the "Hide Called Functions" popup menu
 * action.
 */
void GraphWidget::slotHideCalled()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Remove edges and redraw the graph
	removeEdges(pNode, true);
	draw();
}

/**
 * Shows functions calling tothe current function node.
 * This slot is connected to the "Show Calling Functions" popup menu
 * action.
 */
void GraphWidget::slotShowCalling()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Run a query for called functions
	m_sQueriedFunc = pNode->getText();
	m_bCalled = false;
	m_pCscope->query(CscopeFrontend::Calling, m_sQueriedFunc, 
		Config().getGraphMaxNodeDegree());
}

/**
 * Shows a list of function calls to the current node.
 * The list is displayed in a query dialogue. The user can the select which
 * calls should be displayed in the graph.
 * This slot is connected to the "List Calling Functions" popup menu
 * action.
 */
void GraphWidget::slotListCalling()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	QueryViewDlg dlg;
	
	// Show the query view dialogue
	dlg.query(CscopeFrontend::Calling, pNode->getText());
	if (dlg.exec() != QDialog::Accepted)
		return;
		
	// The OK button was clicked, replace current calls with the listed ones
	removeEdges(pNode, false);
	
	QueryView::Iterator itr;
	CallData data;
	
	data.m_sCallee = pNode->getText();
	
	// Add all listed calls
	for (itr = dlg.getIterator(); !itr.isEOF(); itr.next()) {
		data.m_sCaller = itr.getFunc();
		data.m_sFile = itr.getFile();
		data.m_sLine = itr.getLine();
		data.m_sText = itr.getText();
		
		addCall(data);
	}
	
	// Redraw the graph
	draw();
}

/**
 * Hides functions calling to the current function node.
 * This slot is connected to the "Hide CallingFunctions" popup menu
 * action.
 */
void GraphWidget::slotHideCalling()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Remove edges and redraw the graph
	removeEdges(pNode, false);
	draw();
}

/**
 * Looks up the definition of the current function node.
 * This slot is connected to the "Find Definition" popup menu action.
 */
void GraphWidget::slotFindDef()
{
	GraphNode* pNode;
	QueryViewDlg* pDlg;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Create a query view dialogue
	pDlg = new QueryViewDlg(QueryViewDlg::DestroyOnSelect, this);
	
	// Display a line when it is selected in the dialogue
	connect(pDlg, SIGNAL(lineRequested(const QString&, uint)), this,
		SIGNAL(lineRequested(const QString&, uint)));
		
	// Start the query
	pDlg->query(CscopeFrontend::Definition, pNode->getText());
}

/**
 * Deletes a node from the graph (alogn with all edges connected to this 
 * node).
 * The node removed is the one over which the context menu was invoked.
 * This slot is connected to the "Remove" popup menu action.
 */
void GraphWidget::slotRemoveNode()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
		
	// Delete the node and redraw the graph.
	agdelete(m_pGraph, pNode->getNode());
	draw();
}

/**
 * Shows the list of calls that is represented by a single multi-call node.
 * This slot handles the "Details..." command of the multi-call node menu.
 */
void GraphWidget::slotMultiCallDetails()
{
	GraphNode* pNode;
	QString sFunc;
	uint nType;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Get the required information from the node
	sFunc = GET_NODE_ATTR(pNode->getNode(), "kscope_multi_parent");
	nType = IS_TRUE(GET_NODE_ATTR(pNode->getNode(), "kscope_multi_called")) ?
		CscopeFrontend::Called : CscopeFrontend::Calling;
		
	QueryViewDlg dlg;
	
	dlg.query(nType, sFunc);
	dlg.exec();
}

/**
 * Emits a signal to open an editor at the file and line matching the call
 * information of the current edge.
 * This slot is connected to the "Open Call" popup menu action (for edges).
 */
void GraphWidget::slotOpenCall()
{
	GraphEdge* pEdge;
	QString sFile, sLine;
	
	// Make sure the menu item is an edge
	pEdge = dynamic_cast<GraphEdge*>(m_pMenuItem);
	if (pEdge == NULL)
		return;
	
	// Get call information from the edge
	sFile = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_file");
	sLine = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_line");
	
	// Emit the signal
	emit lineRequested(sFile, sLine.toUInt());
}

#include "graphwidget.moc"
